// this file is covered by the General Public License version 2 or later
// please see GPL.html for more details and licensing issues
// copyright Etienne de Foras ( the author )  mailto: etienne.deforas@gmail.com

#include <cassert>

#include "Device.h"
#include "SurfaceManager.h"
#include "Surface.h"
#include "Light.h"

//////////////////////////////////////////////////////////////////////////////
Device::Device()
{
    _pLight=new Light;
    _dTiltX=0.;
    _dTiltY=0.;
    _iGridX=30;
    _iGridY=30;

    _dAutofocusZMin=0;
    _dAutofocusZMax=0;
    _dAutofocusZStep=0;
}
//////////////////////////////////////////////////////////////////////////////
Device::~Device()
{
    delete _pLight;

    for (unsigned int iPos=0;iPos<_theSurf.size();iPos++)
    {
        SurfaceManager::destroy(_theSurf[iPos]);
    }
}
//////////////////////////////////////////////////////////////////////////////
void Device::insert_surface(string sGeometry,string sKind,int iPos)
{
    assert(iPos>=0);
    assert(iPos<=(int)_theSurf.size()); //accept to insert at the end

    Surface* pSurf=SurfaceManager::create(sGeometry);
    assert(pSurf!=0);
    pSurf->set_kind(sKind);

    pSurf->set("diameter",0.);
    pSurf->set("conic",0.);
    pSurf->set("radius_curvature",9e99);
    pSurf->set("z",0.);

    _theSurf.insert(_theSurf.begin()+iPos,pSurf);
}
//////////////////////////////////////////////////////////////////////////////
void Device::delete_surface(int iPos)
{
    assert(iPos>=0);
    assert(iPos<(int)_theSurf.size());

    _theSurf.erase(_theSurf.begin()+iPos);
}
//////////////////////////////////////////////////////////////////////////////
int Device::nb_surface() const
{
    return _theSurf.size();
}
//////////////////////////////////////////////////////////////////////////////
string Device::get_kind(int iSurf) const
{
    assert(iSurf<nb_surface());
    return _theSurf[iSurf]->kind();
}
//////////////////////////////////////////////////////////////////////////////
bool Device::set_kind(int iSurf,string sKind)
{
    assert(iSurf<nb_surface());
    return _theSurf[iSurf]->set_kind(sKind);
}
//////////////////////////////////////////////////////////////////////////////
bool Device::set_geometry(int iSurf,string sGeometry)
{
    assert(iSurf<nb_surface());
    Surface* pOld=_theSurf[iSurf];
    Surface* pNew=SurfaceManager::create(sGeometry);

    double dZ;
    pOld->get("z",dZ);
    pNew->set("z",dZ);

    double dRC;
    pOld->get("radius_curvature",dRC);
    pNew->set("radius_curvature",dRC);

    double dDiameter;
    pOld->get("diameter",dDiameter);
    pNew->set("diameter",dDiameter);

    double dConic;
    pOld->get("conic",dConic);
    pNew->set("conic",dConic);

    string sKind=pOld->kind();
    pNew->set_kind(sKind);

    _theSurf[iSurf]=pNew;
    delete pOld;
    return true;
}
//////////////////////////////////////////////////////////////////////////////
string Device::get_geometry(int iSurf) const
{
    assert(iSurf<nb_surface());
    return _theSurf[iSurf]->geometry();
}
//////////////////////////////////////////////////////////////////////////////
bool Device::get(int iSurface,string sParam,double& dVal)
{
    assert(iSurface>=0);
    assert(iSurface<nb_surface());

    ray_trace(iSurface);
    return _theSurf[iSurface]->get(sParam,dVal);
}
//////////////////////////////////////////////////////////////////////////////
bool Device::set(int iSurface,string sParam,double dVal)
{
    assert(iSurface>=0);
    assert(iSurface<nb_surface());

    return _theSurf[iSurface]->set(sParam,dVal);
}
//////////////////////////////////////////////////////////////////////////////
void Device::ray_trace(int iSurfaceMax)
{
    if (nb_surface()==0)
        return;

    initialise_light();
    for (int i=0;i<=iSurfaceMax;i++)
    {
        receive(*_pLight,i);
    }
}
//////////////////////////////////////////////////////////////////////////////
void Device::set_light_grid(int iGridX,int iGridY)
{
    _iGridX=iGridX;
    _iGridY=iGridY;
}
//////////////////////////////////////////////////////////////////////////////
Light& Device::get_light_at(int iSurface)
{
    ray_trace(iSurface);
    return *_pLight;
}
//////////////////////////////////////////////////////////////////////////////
double Device::get_profile_at(int iSurface,double dX,double dY)
{
    //  ray_trace(iSurface);
    return _theSurf[iSurface]->get_profile(dX,dY);
}
//////////////////////////////////////////////////////////////////////////////
void Device::initialise_light()
{
    Surface* pSurf=_theSurf[0];
    _pLight->set_nb_photons(_iGridX,_iGridY);
    _pLight->set_tilt(_dTiltX,_dTiltY);
    _pLight->set_colors(_sLightColors);
    _pLight->set_geometry(pSurf->z(),pSurf->diameter());
    _pLight->nb_photon(); //to initialise it
}
//////////////////////////////////////////////////////////////////////////////
void Device::receive(Light& l,int iSurface)
{
    _theSurf[iSurface]->receive(l);
}
//////////////////////////////////////////////////////////////////////////////
void Device::set_light_tilt(double dTiltX,double dTiltY)
{
    _dTiltX=dTiltX;
    _dTiltY=dTiltY;
}
//////////////////////////////////////////////////////////////////////////////
void Device::get_light_tilt(double& dTiltX,double& dTiltY) const
{
    dTiltX=_dTiltX;
    dTiltY=_dTiltY;
}
//////////////////////////////////////////////////////////////////////////////
void Device::set_light_colors(const string & sLightColors)
{
    _sLightColors=sLightColors;
}
//////////////////////////////////////////////////////////////////////////////
void Device::get_light_colors(string& sLightColors) const
{
    sLightColors=_sLightColors;
}
//////////////////////////////////////////////////////////////////////////////
void Device::set_autofocus(double dAutofocusZMin,double dAutofocusZMax,double dAutofocusZStep)
{
    if (dAutofocusZStep<1e-9)
        dAutofocusZStep=1e-9;
    _dAutofocusZStep=dAutofocusZStep;

    if (dAutofocusZMin<dAutofocusZMax)
    {
        _dAutofocusZMin=dAutofocusZMin;
        _dAutofocusZMax=dAutofocusZMax;
    }
    else
    {
        _dAutofocusZMin=dAutofocusZMax;
        _dAutofocusZMax=dAutofocusZMin;
    }
}
//////////////////////////////////////////////////////////////////////////////
void Device::get_autofocus(double& dAutofocusZMin,double& dAutofocusZMax,double& dAutofocusZStep)
{
    dAutofocusZMin=_dAutofocusZMin;
    dAutofocusZMax=_dAutofocusZMax;
    dAutofocusZStep=_dAutofocusZStep;
}
//////////////////////////////////////////////////////////////////////////////
void Device::enable_auto_diameter(bool bEnable)
{
    //TODO
}
//////////////////////////////////////////////////////////////////////////////
