// this file is covered by the General Public License version 2 or later
// please see GPL.html for more details and licensing issues
// copyright Etienne de Foras ( the author )  mailto: etienne.deforas@gmail.com

#include "FrameSideView.h"
#include "ui_FrameSideView.h"
#include "QtLensDefines.h"
#include <QtGui>
#include <cmath>
#include <cassert>

#include "Device.h"
#include "Light.h"
#include "Photon.h"

void FrameSideView::enlarge(QRectF& r,double dRatio)
{
    double dX=r.width();
    double dY=r.height();
    QPointF c=r.center();
    r.setCoords( c.x()-dX*dRatio/2., c.y()-dY*dRatio/2., c.x()+dX*dRatio/2., c.y()+dY*dRatio/2.);
}


class myScene : public QGraphicsScene
{
protected:
    virtual void wheelEvent ( QGraphicsSceneWheelEvent * wheelEvent )
    {
        QGraphicsView* v=views()[0];
        v->setTransformationAnchor(QGraphicsView::AnchorUnderMouse);

        if (wheelEvent->delta()>0)
            v->scale(1.25,1.25);
        else
            v->scale(0.8,0.8);

        wheelEvent->accept();
    }
};

FrameSideView::FrameSideView(QWidget *parent) :
        QFrame(parent),
        m_ui(new Ui::FrameSideView)
{
    m_ui->setupUi(this);
    m_ui->graphicsView->setScene(new myScene); //TODO deleter la scene?

    m_ui->graphicsView->setDragMode (QGraphicsView::ScrollHandDrag);
    m_ui->graphicsView->setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
    m_ui->graphicsView->setVerticalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
    m_ui->graphicsView->setRenderHints( QPainter::Antialiasing );

    _pDevice=0;
}

FrameSideView::~FrameSideView()
{
    delete m_ui;
}

void FrameSideView::changeEvent(QEvent *e)
{
    QFrame::changeEvent(e);
    switch (e->type())
    {
    case QEvent::LanguageChange:
        m_ui->retranslateUi(this);
        break;
    default:
        break;
    }
}

void FrameSideView::device_changed(Device* pDevice)
{
    bool bSameDevice=(_pDevice==pDevice);
    _pDevice=pDevice;
    QGraphicsScene* scene=m_ui->graphicsView->scene();
    scene->clear();

    if (pDevice->nb_surface()==0)
        return;

        qDebug("FrameSideView::device_changed and raytrace");
    _pDevice->enable_auto_diameter(false);

    vector<double> viX,viY;
    int iLastSolidColor=0xffffff;
    for (int iSurf=0;iSurf<_pDevice->nb_surface();iSurf++)
    {
        double dZSurf,dDiameter;
        _pDevice->get(iSurf,"z",dZSurf);
        _pDevice->get(iSurf,"diameter",dDiameter);

        vector<double> viXNew,viYNew;
        for (int iStep=-SIDEVIEW_NB_POINTS_SURFACE;iStep<=SIDEVIEW_NB_POINTS_SURFACE;iStep++)
        {
            double dDiam=dDiameter/2.*iStep/(double)SIDEVIEW_NB_POINTS_SURFACE;
            viXNew.push_back(dDiam);
            double dZ=_pDevice->get_profile_at(iSurf,0.,dDiam);
            assert(!isnan(dZ));
            viYNew.push_back(dZ+dZSurf);
        }

        if(iLastSolidColor==0xffffff)
        { //disp as white so just a line
            for (unsigned int i=0;i<viXNew.size()-1;i++)
            {
                scene->addLine(viYNew[i],viXNew[i],viYNew[i+1],viXNew[i+1])->setZValue(10);
                assert(!isnan(viYNew[i]));
                assert(!isnan(viXNew[i]));
                assert(!isnan(viYNew[i+1]));
                assert(!isnan(viXNew[i+1]));
            }
        }
        else
        {
            //draw a polygon
            QPolygonF polygon;
            for (unsigned int i=0;i<viXNew.size();i++)
            {
                assert(!isnan(viYNew[i]));
                assert(!isnan(viXNew[i]));

                polygon << QPointF(viYNew[i],viXNew[i]);
            }
            for (int i=viX.size()-1;i>=0;i--)
            {
                assert(!isnan(viY[i]));
                assert(!isnan(viX[i]));

                polygon << QPointF(viY[i],viX[i]);
            }

            scene->addPolygon(polygon,QPen(),QBrush(QColor(QRgb(iLastSolidColor))))->setZValue(10);
        }

        double dSolidColor;
        _pDevice->get(iSurf,"solid_color",dSolidColor);
        iLastSolidColor=(int)dSolidColor;

        viX=viXNew;
        viY=viYNew;
    }
    //return;
    _pDevice->set_light_grid(SIDEVIEW_NB_POINTS_LIGHT,1);

    vector<Photon> vp1=_pDevice->get_light_at(-1).get_all_photons();
    vector<Photon> vp2;
    for (int iS=0;iS<_pDevice->nb_surface();iS++)
    {
        vp2=_pDevice->get_light_at(iS).get_all_photons();
        for (unsigned int iP=0;iP<vp2.size();iP++)
        {
            Photon& p1=vp1[iP];
            Photon& p2=vp2[iP];

            if (p1.is_valid() & p2.is_valid())
            {
                scene->addLine(p1.z,p1.x,p2.z,p2.x,QPen(QRgb(p1.iVisualColor)))->setZValue(20);
            }
            else if (p1.is_valid() &!p2.is_valid() )
            {
                QPen qp(QRgb(p1.iVisualColor));
                qp.setStyle(Qt::DotLine);

                if((p1.z-p2.z)*(p1.z-p2.z)+(p1.x-p2.x)*(p1.x-p2.x)>1e-16) //TODO reporter ce bug a Qt
                    scene->addLine(p1.z,p1.x,p2.z,p2.x,qp)->setZValue(20);
            }
        }

        vp1=vp2;
    }
    //return;
    if (bSameDevice==false)
    {
        QRectF r=scene->itemsBoundingRect();
        enlarge(r,1.1);
        scene->setSceneRect(r);
        m_ui->graphicsView->fitInView(r,Qt::KeepAspectRatio);
    }
    else
    {
        QRectF r=scene->sceneRect();
        QRectF r2=scene->itemsBoundingRect();
        enlarge(r2,1.1);
        scene->setSceneRect(r.united(r2));
    }
}
