// this file is covered by the General Public License version 2 or later
// please see GPL.html for more details and licensing issues
// copyright Etienne de Foras ( the author )  mailto: etienne.deforas@gmail.com

#include <cassert>
#include <cmath>
#include <algorithm>

#include "Light.h"
#include "Photon.h"
#include "MediumManager.h"

#define PI (3.1415926535897)
#define RAD_TO_DEG (360./2./PI)
#define DEG_TO_RAD (PI*2./360.)

//////////////////////////////////////////////////////////////////////////////
Light::Light(int iDimX,int iDimY)
{
    set_nb_photons(iDimX,iDimY);
    _iNbPhotons=0;
    _dTiltX=0.;
    _dTiltY=0.;
    _dPsfDiameter=0.;

    _dCenterX=0.;
    _dCenterY=0.;

    _dDiameter=0.;
    _dZ=0.;
    _dFD=0.;

    _bPhotonsReady=false;

    _pFirstMedium=MediumManager::create("air");
    _pMedium=_pFirstMedium;
}
//////////////////////////////////////////////////////////////////////////////
Light::~Light()
{
    MediumManager::destroy(_pFirstMedium);
}
//////////////////////////////////////////////////////////////////////////////
int Light::nb_photon()
{
    if (_bPhotonsReady==false)
        initialise();
    return _iNbPhotons;
}
//////////////////////////////////////////////////////////////////////////////
Photon& Light::get_photon(int iPos)
{
    assert(iPos>=0);
    assert(iPos<_iNbPhotons);

    if (_bPhotonsReady==false)
        initialise();

    return _vPhotons[iPos];
}
//////////////////////////////////////////////////////////////////////////////
vector<Photon> Light::get_all_photons()
{
    return _vPhotons;
}
//////////////////////////////////////////////////////////////////////////////
void Light::set_geometry(double dZ,double dDiameter)
{
    assert(dDiameter!=-1.);
    _dZ=dZ;
    _dDiameter=dDiameter;
    _bPhotonsReady=false;
}
//////////////////////////////////////////////////////////////////////////////
void Light::initialise()
{
    //corrige la direction avec le tilt
    double dTanX=tan(_dTiltX);
    double dTanY=tan(_dTiltY);

    // normalise la direction; dNorm!=0 par def
    double dNorm=sqrt(dTanX*dTanX+dTanY*dTanY+1.*1.);
    double dTanXN=dTanX/dNorm;
    double dTanYN=dTanY/dNorm;
    double dIdentN=-1./dNorm;;

    vector<double> vdLambda;
    vector<int> viVisualColors;
    calc_colors(_sAllColors, vdLambda, viVisualColors);

    _iNbPhotons=_iDimX*_iDimY*vdLambda.size();
    _vPhotons.resize(_iNbPhotons); //TODO optimiser les alloc

    int iPhoton=0;

    for (unsigned int iLambda=0;iLambda<vdLambda.size();iLambda++)
    {
        double dLambda=vdLambda[iLambda];
        int iVisualColor=viVisualColors[iLambda];

        for (int j=0; j<_iDimY;j++)
        {
            double dPosY=-_dDiameter/2.+_dDiameter*j/_iDimY;
            if (_iDimY==1)
                dPosY=0;

            for (int i=0;i<_iDimX;i++)
            {
                Photon& p=_vPhotons[iPhoton];
                p.lambda=dLambda;
                p.iVisualColor=iVisualColor;

                // a optimiser au besoin
                double dPosX=-_dDiameter/2.+_dDiameter*i/_iDimX;
                /*
                if ( !_pSurf->verify_sq_radius(dPosX*dPosX+dPosY*dPosY) ) //TODO
                {
                    p.valid=false;
                    continue;
                }
                */
                p.x=dPosX;
                p.y=dPosY;
                p.z=_dZ+_dDiameter/2.; //recul
                p.dx=dTanXN;
                p.dy=dTanYN;
                p.dz=dIdentN;

                p.valid=true;
                iPhoton++;
            }
        }
    }

    _pMedium=_pFirstMedium;
    _bPhotonsReady=true;
}
//////////////////////////////////////////////////////////////////////////////
void Light::set_tilt(double dTiltX,double dTiltY)
{
    _dTiltX=dTiltX*DEG_TO_RAD;
    _dTiltY=dTiltY*DEG_TO_RAD;
    _bPhotonsReady=false;
}
//////////////////////////////////////////////////////////////////////////////
void Light::get_tilt(double& dTiltX,double& dTiltY)
{
    dTiltX=_dTiltX*RAD_TO_DEG;
    dTiltY=_dTiltY*RAD_TO_DEG;
}
//////////////////////////////////////////////////////////////////////////////
void Light::calc_colors(string sColors,vector<double>& vdLambda,vector<int>& vdVisualColors)
{
    vdLambda.clear();
    vdVisualColors.clear();

    if (sColors.find("IR.")!=string::npos)
    {
        vdLambda.push_back(1000e-6);
        vdVisualColors.push_back(0x74572e); //brown
    }

    if (sColors.find("Red.")!=string::npos)
    {
        vdLambda.push_back(656.2725e-6);
        vdVisualColors.push_back(0xff0000);
    }

    if (sColors.find("YellowBlack.")!=string::npos)
    {
        vdLambda.push_back(587.5618e-6);
        vdVisualColors.push_back(0x000000);
    }

    if (sColors.find("Yellow.")!=string::npos)
    {
        vdLambda.push_back(587.5618e-6);
        vdVisualColors.push_back(0xede500); //yellow
    }

    if (sColors.find("Green.")!=string::npos)
    {
        vdLambda.push_back(546.0740e-6);
        vdVisualColors.push_back(0xff00);
    }

    if (sColors.find("Blue.")!=string::npos)
    {
        vdLambda.push_back(486.1327e-6);
        vdVisualColors.push_back(0xff);
    }

    if (sColors.find("UV.")!=string::npos)
    {
        vdLambda.push_back(300.e-6);
        vdVisualColors.push_back(0x772176); //violet
    }
}
//////////////////////////////////////////////////////////////////////////////
void Light::set_colors(const string & sAllColors)
{
    _sAllColors=sAllColors;

    if (_sAllColors.empty())
        _sAllColors="YellowBlack.";

    _bPhotonsReady=false;
}
//////////////////////////////////////////////////////////////////////////////
void Light::set_nb_photons(int iNbX,int iNbY)
{
    assert(iNbX>0);
    assert(iNbY>0);

    _iDimX=iNbX;
    _iDimY=iNbY;
    //_iNbPhotons=0;
    _bPhotonsReady=false;
}
//////////////////////////////////////////////////////////////////////////////
Medium& Light::medium()
{
    return *_pMedium;
}
//////////////////////////////////////////////////////////////////////////////
void Light::medium(Medium* pM)
{
    assert(pM!=0);
    _pMedium=pM;
}
//////////////////////////////////////////////////////////////////////////////
void Light::get_psf_center(double& dCenterX,double& dCenterY) const
{
    dCenterX=_dCenterX;
    dCenterY=_dCenterY;
}
//////////////////////////////////////////////////////////////////////////////
double Light::get_psf_diameter() const
{
    return _dPsfDiameter;
}
//////////////////////////////////////////////////////////////////////////////
void Light::calc_psf()
{
    double dXS=0., dYS=0.;
    double dMinX=0.,dMaxX=0.,dMinY=0.,dMaxY=0.;
    double tanalpha2=0; //for compiler warning

    int iNbValidPhoton=0;
    for (int i=0;i<_iNbPhotons;i++)
    {
        Photon& p=_vPhotons[i];

        if (p.valid==false)
            continue;

        if (iNbValidPhoton==0)
        {
            dMinX=p.x;
            dMaxX=p.x;
            dMinY=p.y;
            dMaxY=p.y;
            tanalpha2=(p.dx*p.dx+p.dy*p.dy)/(p.dz*p.dz);
        }
        else
        {
            dMinX=std::min(dMinX,p.x);
            dMaxX=std::max(dMaxX,p.x);
            dMinY=std::min(dMinY,p.y);
            dMaxY=std::max(dMaxY,p.y);
            // calcul de F/D: rapport F/D reel, cherche les faisceaux les plus hors axes.
            // pour l'instant calcule juste sur l'axe
            double tanalpha2temp=(p.dx*p.dx+p.dy*p.dy)/(p.dz*p.dz);
            if (tanalpha2temp>tanalpha2)
            {
                tanalpha2=tanalpha2temp;
            }
        }

        dXS+=p.x;
        dYS+=p.y;

        iNbValidPhoton++;

        // calcul de F/D: rapport F/D reel, cherche les faisceaux les plus hors axes.
        // pour l'instant calcule juste sur l'axe, TODO
        double tanalpha2temp=(p.dx*p.dx+p.dy*p.dy)/(p.dz*p.dz);
        if (tanalpha2temp>tanalpha2)
        {
            tanalpha2=tanalpha2temp;
        }
    }

    if(iNbValidPhoton==0)
    {
        _dCenterX=0.;
        _dCenterY=0.;
        _dPsfDiameter=0.;
        _dFD=0.;
        return;
    }

    _dCenterX=dXS/iNbValidPhoton;
    _dCenterY=dYS/iNbValidPhoton;

    _dPsfDiameter=std::max(dMaxX-dMinX,dMaxY-dMinY);

    // fin du calcul du rapport F/D
    if (tanalpha2>0.)
        _dFD=0.5/sqrt(tanalpha2);
    else
        _dFD=-1.;

}
//////////////////////////////////////////////////////////////////////////////
double Light::get_FD() const
{
    return _dFD;
}
//////////////////////////////////////////////////////////////////////////////
double Light::get_lfro() const
{
    return get_psf_diameter()/2./get_theorical_ro();
}
//////////////////////////////////////////////////////////////////////////////
double Light::get_theorical_ro() const
{
    return 1.22*get_FD()*560e-6; //lumiere jaune tout en mm TODO
}
//////////////////////////////////////////////////////////////////////////////
