// this file is covered by the General Public License version 2 or later
// please see GPL.html for more details and licensing issues
// copyright Etienne de Foras ( the author )  mailto: etienne.deforas@gmail.com

#include <cassert>
#include <cmath>

#include "Surface.h"

#include "Photon.h"
#include "Light.h"
#include "MediumManager.h"
#include "Medium.h"

//////////////////////////////////////////////////////////////////////////////
Surface::Surface()
{
    diameter(-1.);
    _pMedium=0;
    _pMediumNext=0;
    _pMediumPrev=0;
}
//////////////////////////////////////////////////////////////////////////////
Surface::~Surface()
{
    delete _pMedium;
}
//////////////////////////////////////////////////////////////////////////////
void Surface::diameter(double dDiameter)
{
    _dDiameter=dDiameter;
    _bDiameterAuto=(_dDiameter==-1.);
    if (_bDiameterAuto)
        _dDiameter2=-1.;
    else
        _dDiameter2=_dDiameter*_dDiameter;
}
//////////////////////////////////////////////////////////////////////////////
double Surface::diameter()
{
    if (_dDiameter==-1.)
    {
        if (_dDiameter2==-1.)
            return -1.;
        _dDiameter=sqrt(_dDiameter2); // et les valeurs negatives?
    }
    return _dDiameter;
}
//////////////////////////////////////////////////////////////////////////////
void Surface::z(double dz)
{
    _z=dz;
}
//////////////////////////////////////////////////////////////////////////////
double Surface::z()
{
    return _z;
}
//////////////////////////////////////////////////////////////////////////////
bool Surface::verify_sq_radius(double d)
{
    double diam=4.*d;

    // pour le diametre max
    if (_bDiameterAuto)
    {
        if ( (_dDiameter2==-1.) || (diam>_dDiameter2) ) // lappalisade...
        {
            _dDiameter2=diam;
            _dDiameter=-1.;
        }
    }
    else
    {
        if (diam>_dDiameter2)
            return false;
    }

    return true;
}
//////////////////////////////////////////////////////////////////////////////
void Surface::local_ref(Photon& p)
{
    // translation simple de ref, en attendant le besoins de gerer + de choses
    p.z-=z();
}
//////////////////////////////////////////////////////////////////////////////
string Surface::kind()
{
    return _sKind;
}
//////////////////////////////////////////////////////////////////////////////
bool Surface::set_kind(string sKind)
{
    assert(sKind!="");
    if ( (sKind!="reflect") && (sKind!="stop") && (sKind!="observe") && (sKind!="transmit"))
    {
        MediumManager::destroy(_pMedium);
        _pMedium=MediumManager::create(sKind);
        assert(_pMedium!=0);
    }

    _sKind=sKind;
    return true;
}
//////////////////////////////////////////////////////////////////////////////
string Surface::geometry()
{
    return _sGeometry;
}
//////////////////////////////////////////////////////////////////////////////
void Surface::global_ref(Photon& p)
{
    // translation simple de ref, en attendant le besoins de gerer + de choses
    p.z+=z();
}
//////////////////////////////////////////////////////////////////////////////
bool Surface::get(string sParam,double& dVal)
{
    if (sParam=="diameter")
    {
        dVal=diameter();
        return true;
    }

    if (sParam=="solid_color")
    {
        if(_pMedium==0)
            dVal=0xffffff;
        else
            dVal=_pMedium->solid_color();
        return true;
    }

    if (sParam=="z")
    {
        dVal=_z;
        return true;
    }

    return false;
}
//////////////////////////////////////////////////////////////////////////////
bool Surface::set(string sParam,double dVal)
{
    if (sParam=="diameter")
    {
        diameter(dVal);
        return true;
    }

    if (sParam=="z")
    {
        z(dVal);
        return true;
    }

    return false;
}
//////////////////////////////////////////////////////////////////////////////
void Surface::receive(Light& l)
{
    assert(_sKind!="");
    if (_sKind=="reflect")
        reflect(l);
    else if (_sKind=="stop")
        stop(l);
    else if (_sKind=="observe")
        observe(l);
    else
        transmit(l);
}
//////////////////////////////////////////////////////////////////////////////
void Surface::observe(Light& l)
{
    stop(l);
    l.calc_psf();
}
//////////////////////////////////////////////////////////////////////////////
// arrete les photons
void Surface::stop(Light& l)
{
    for (int i=0;i<l.nb_photon();i++)
    {
        Photon& p=l.get_photon(i);

        if (!p.is_valid())
            continue;

        local_ref(p);

        stop_photon(p);
        global_ref(p);

        if (!p.is_valid())
            continue;
    }
}
///////////////////////////////////////////////////////////////////////////////
// reflechis les photons
void Surface::reflect(Light& l)
{
    for (int i=0;i<l.nb_photon();i++)
    {
        Photon& p=l.get_photon(i);

        if (!p.is_valid())
            continue;

        local_ref(p);

        reflect_photon(p);
        if (!p.is_valid())
            continue;

        global_ref(p);
    }
}
///////////////////////////////////////////////////////////////////////////////
void Surface::transmit(Light& l)
{
    _pMediumPrev=&(l.medium());
    _pMediumNext=_pMedium;

    assert(_pMediumPrev);

    for (int i=0;i<l.nb_photon();i++)
    {
        Photon& p=l.get_photon(i);

        if (!p.is_valid())
            continue;

        local_ref(p);

        transmit_photon(p);
        if (!p.is_valid())
            continue;

        global_ref(p);
    }

    if(_pMedium!=0)
        l.medium(_pMedium);
}
//////////////////////////////////////////////////////////////////////////////
double Surface::get_profile(double dX,double dY)
{
    Photon p;
    p.x=dX;
    p.y=dY;
    p.z=_dDiameter*2.; //TODO

    p.dx=0.;
    p.dy=0.;
    p.dz=-1.;
    p.valid=true;

    stop_photon(p,false);

    if(p.is_valid()==false)
        return 0;

    return p.z;
}
//////////////////////////////////////////////////////////////////////////////
