// this file is covered by the General Public License version 2 or later
// please see GPL.html for more details and licensing issues
// copyright Etienne de Foras ( the author )  mailto: etienne.deforas@gmail.com

#include <iostream>
#include <cmath>
#include <cassert>

#include "SurfaceConic.h"
#include "Photon.h"
#include "Medium.h"

///////////////////////////////////////////////////////////////////////////////
SurfaceConic::SurfaceConic()
{
    _dConic=0.;
    _dCurvature=0.;
    _sGeometry="conic";
}
//////////////////////////////////////////////////////////////////////////////
bool SurfaceConic::get(string sParam,double& dVal)
{
    if (sParam=="conic")
    {
        dVal=_dConic;
        return true;
    }

    if (sParam=="curvature")
    {
        dVal=_dCurvature;
        return true;
    }

    if (sParam=="radius_curvature")
    {
        dVal=1./_dCurvature;
        return true;
    }  // on espere !=0

    return Surface::get(sParam,dVal);
}
//////////////////////////////////////////////////////////////////////////////
bool SurfaceConic::set(string sParam,double dVal)
{
    if (sParam=="conic")
    {
        _dConic=dVal;
        return true;
    }

    if (sParam=="curvature")
    {
        _dCurvature=dVal;
        return true;
    }

    if (sParam=="radius_curvature")
    {
        _dCurvature=1./dVal;
        return true;
    }

    return Surface::set(sParam,dVal);
}
///////////////////////////////////////////////////////////////////////////////
void SurfaceConic::stop_photon(Photon& p,bool bVerifyRadius)
{
    if (p.is_valid()==false)
        return;

    if(isinf(_dCurvature))
    {
        p.valid=false;
        return;
    }

    //on calcule les coef de l'eq du 2eme degree en t:
    //the equation is t^2A+tB+C=0

    if (p.dz==0.)
    {
        p.valid=false; // pour l'instant, je ne sait que faire!
        return;
    }

    // arrete le photon en z=0
    double tmin=-p.z/p.dz;
    p.x+=tmin*p.dx;
    p.y+=tmin*p.dy;
    p.z=0.;

    // formules simplifiees avec p.z=0.
    double dA=_dCurvature*(p.dx*p.dx+p.dy*p.dy+(_dConic+1.)*p.dz*p.dz);
    double dB=2.*(_dCurvature*(p.x*p.dx+p.y*p.dy)-p.dz);
    double dC=_dCurvature*(p.x*p.x+p.y*p.y);

    double tfinal;

    if (dA==0.)
    {
        //la trajectoire est verticale!
        if (dB==0.)
        {
            p.valid=false;
            return;
        }

        //the equation is now : t*dB+dC=0 so:
        tfinal=-dC/dB;
    }
    else //dA!=0
    {
        //on resoud l'equation
        double dB2=dB*dB;
        double delta=dB2-4.*dC*dA;

        if (delta<0.)
        {
            p.valid=false;
            return;
        }

        double sqrtDelta=sqrt(delta);
        double t1,t2;

        if (delta!=dB2) // alors les denum existent (test a parfaire)
        {
            t1=(2.*dC)/(+sqrtDelta-dB);
            t2=(2.*dC)/(-sqrtDelta-dB);
        }
        else
        {
            // ici delta~=dB2
            // on utilise une solution approchee:
            if (dB!=0)
            {
                t1=-dC/dB;
                t2=10.*t1; // pour etre sur de choisir t1
            }
            else
            {
                p.valid=false;  // on ne devrait jamais passer ici !??!?
                return;
            }
        }

        // on prend le t qui donne le z le plus proche de zero
        //t1ok=abs(z+t1.*dz)<abs(z+t2.*dz); //oct ver
        if (t1*t1<t2*t2) //a optimiser plus tard...
            tfinal=t1;
        else
            tfinal=t2;
    }

    //verifie que la surface est bien dans le futur du photon
    if ( tmin+tfinal<0 )
    {
        p.valid=false;
        return;
    }

    p.x+=p.dx*tfinal;
    p.y+=p.dy*tfinal;
    p.z+=p.dz*tfinal;

    assert(!isnan(p.x));
    assert(!isnan(p.y));
    assert(!isnan(p.z));
    assert(!isnan(p.dy));
    assert(!isnan(p.dz));

    if (bVerifyRadius==false)
        return;

    double dRadiusSq=p.x*p.x+p.y*p.y;
    if ( verify_sq_radius(dRadiusSq))
        return; // tout est ok

    //pas d'intersection valable, on marque le point et on sort!
    p.valid=false;
}
///////////////////////////////////////////////////////////////////////////////
void SurfaceConic::reflect_photon(Photon &p)
{
    if (!p.is_valid())
        return;

    stop_photon(p);
    if (!p.is_valid())
        return;

    //calcule la normale en ce point ( de distance au centre: ro)
    double dRadiusSq=p.x*p.x+p.y*p.y;

    if (dRadiusSq>0.)
    {
        // le vecteur rayon, tangent a la surface
        //double denom=(_dCurvature*dRadiusSq-p.z);
        double dTangRadius=p.z*(2.+p.z*p.z*(_dConic+1.)*_dCurvature/(_dCurvature*dRadiusSq-p.z));

        // le vecteur tangent au BORD de la surface donc parallele a la circonference
        // produit vectoriel pour calculer la normale, a encapsuler plus tard
        double nx=-p.x*dTangRadius;
        double ny=-p.y*dTangRadius;
        double nz=dRadiusSq;

        // a partir d'ici, on calcule avec la normale (nx,ny,nz)
        double dRayonSq=nx*nx+ny*ny+nz*nz;
        assert(dRayonSq>0.);
        double u=2.*(nx*p.dx+ny*p.dy+nz*p.dz)/dRayonSq;
        p.dx-=nx*u;
        p.dy-=ny*u;
        p.dz-=nz*u;
    }
    else
    {
        // cas special, on est au centre du miroir;
        p.dz=-p.dz;
    }

    assert(!isnan(p.x));
    assert(!isnan(p.y));
    assert(!isnan(p.z));
    assert(!isnan(p.dy));
    assert(!isnan(p.dz));
}
//////////////////////////////////////////////////////////////////////////////
void SurfaceConic::transmit_photon(Photon& p)
{
    if (!p.is_valid())
        return;

    stop_photon(p);
    if (!p.is_valid())
        return;

    double nx,ny,nz;

    //calcule la normale en ce point ( de distance au centre: ro)
    double dRadiusSq=p.x*p.x+p.y*p.y;
    double denum=_dCurvature*dRadiusSq-p.z;
    if (denum!=0.)
    {
        // le vecteur rayon, tangent a la surface
        double dTangRadius=p.z*(2.+p.z*p.z*(_dConic+1.)*_dCurvature/denum);
        // le vecteur tangent au BORD de la surface donc parallele a la circonference
        // produit vectoriel pour calculer la normale, a encapsuler plus tard
        nx=-p.x*dTangRadius;
        ny=-p.y*dTangRadius;
        nz=dRadiusSq;
    }
    else // au centre du miroir ou surface plate
    {
        nx=0.;
        ny=0.;
        nz=1.;
    }

    // normalize la normale
    double dRayon=sqrt(nx*nx+ny*ny+nz*nz);
    assert(dRayon>0.);
    nx/=dRayon;
    ny/=dRayon;
    nz/=dRayon;

    double u=nx*p.dx+ny*p.dy+nz*p.dz;

    // a partir d'ici, on calcule avec la normale (dx,dy,dz)
    // u est la projection de I sur la normale

    //calcule de C1=u/I
    double dC12=u*u/(p.dx*p.dx+p.dy*p.dy+p.dz*p.dz);

    assert(_pMediumNext!=0);
    assert(_pMediumPrev!=0);
    double dRatioN=_pMediumNext->index(p.lambda)/_pMediumPrev->index(p.lambda);

    double denom=dRatioN*dRatioN+dC12-1.;

    if (denom<=0.)
    {  // reflexion totale
        p.valid=false;
        return;
    }

    double dT2T1=sqrt(dC12/denom);

    p.dx=u*nx*(1.-dT2T1)+dT2T1*p.dx;
    p.dy=u*ny*(1.-dT2T1)+dT2T1*p.dy;
    p.dz=u*nz*(1.-dT2T1)+dT2T1*p.dz;
}
//////////////////////////////////////////////////////////////////////////////
