// this file is covered by the General Public License version 2 or later
// please see GPL.html for more details and licensing issues
// copyright Etienne de Foras ( the author )  mailto: etienne.deforas@gmail.com

#include "mainwindow.h"
#include "ui_mainwindow.h"
#include <QMessageBox>
#include <QFileDialog>
#include <QtGui>

#include <cassert>

#include "Device.h"
#include "DeviceIo.h"
#include "DockSurfacesData.h"
#include "DockLightProperties.h"
#include "DockScatterPlot.h"
#include "DockAutofocus.h"
#include "FrameSideView.h"
#include "QtLensDefines.h"

MainWindow::MainWindow(QWidget *parent)
    : QMainWindow(parent), ui(new Ui::MainWindowClass)
{
    ui->setupUi(this);
    _pDevice=new Device;

    _bMustSave=false;

    _pFrameSideView=new FrameSideView(this);
    connect(_pFrameSideView,SIGNAL(device_changed(void*)),this,SLOT(device_changed(void*)));
    setCentralWidget(_pFrameSideView);

    _pDockSurfacesData=new DockSurfacesData(this);
    addDockWidget(Qt::BottomDockWidgetArea, _pDockSurfacesData);
    connect(_pDockSurfacesData,SIGNAL(device_changed(void*)),this,SLOT(device_changed(void*)));

    _pDockScatterPlot=new DockScatterPlot(this);
    addDockWidget(Qt::LeftDockWidgetArea, _pDockScatterPlot);
    connect(_pDockScatterPlot,SIGNAL(device_changed(void*)),this,SLOT(device_changed(void*)));

    _pDockLightProperties=new DockLightProperties(this);
    addDockWidget(Qt::LeftDockWidgetArea, _pDockLightProperties);
    connect(_pDockLightProperties,SIGNAL(device_changed(void*)),this,SLOT(device_changed(void*)));

    _pDockAutofocus=new DockAutofocus(this);
    addDockWidget(Qt::LeftDockWidgetArea, _pDockAutofocus);
    connect(_pDockAutofocus,SIGNAL(device_changed(void*)),this,SLOT(device_changed(void*)));

    connect(this,SIGNAL(quit()),this,SLOT(on_actionQuit_triggered()));

    device_changed(0);
}

MainWindow::~MainWindow()
{
    delete ui;
    delete _pDevice;
}
//////////////////////////////////////////////////////////////////////////////
void MainWindow::closeEvent(QCloseEvent *event)
{
    if (ask_save_and_action())
        event->accept();
    else
        event->ignore();
}
//////////////////////////////////////////////////////////////////////////////
bool MainWindow::ask_save_and_action()
{
    if (_bMustSave==true)
    {
        int iRet=QMessageBox::question(this,"Warning","Save?",QMessageBox::Yes | QMessageBox::No | QMessageBox::Cancel);
        if (iRet==QMessageBox::Cancel)
            return false;

        if (iRet==QMessageBox::Yes)
        {
            DeviceIo::save(_sFileName,_pDevice);
            _bMustSave=false;
            device_changed(0);
        }
    }

    return true;
}
//////////////////////////////////////////////////////////////////////////////
void MainWindow::clear_device()
{
    _bMustSave=false;
  //  _sFileName="";
    delete _pDevice;
    _pDevice=new Device;

}
//////////////////////////////////////////////////////////////////////////////
void MainWindow::on_actionQuit_triggered()
{
    if (ask_save_and_action())
        qApp->quit();
}
//////////////////////////////////////////////////////////////////////////////
void MainWindow::on_actionAbout_triggered()
{
    QMessageBox mb;
    mb.setText("QtLens"QTLENS_VERSION" by etienne.deforas@gmail.com");
    mb.setIconPixmap(QPixmap(":/QtLens/QtLens.ico"));
    mb.exec();
}
//////////////////////////////////////////////////////////////////////////////
void MainWindow::on_actionLoad_triggered()
{
    if (ask_save_and_action()==false)
        return;

    _sFileName = QFileDialog::getOpenFileName(this,tr("Open QTLens File"), ".", tr("QtLens Files (*.qtlens)")).toStdString();

    if (!_sFileName.empty())
    {
        clear_device();
        device_changed(0);
        load_file(_sFileName);
    }
}
//////////////////////////////////////////////////////////////////////////////
void MainWindow::on_actionNew_triggered()
{
    if (ask_save_and_action()==false)
        return;

    clear_device();
    device_changed(0);
}
//////////////////////////////////////////////////////////////////////////////
void MainWindow::on_actionSave_triggered()
{
    DeviceIo::save(_sFileName,_pDevice);
    _bMustSave=false;
    device_changed(0);
}
//////////////////////////////////////////////////////////////////////////////
void MainWindow::on_actionSave_as_triggered()
{
    if (_pDevice==0)
        return;

    string sFileName = QFileDialog::getSaveFileName(this,tr("Open QTLens File"), ".", tr("QtLens Files (*.qtlens)")).toStdString();

    if (sFileName.empty()==false)
    {
        _sFileName=sFileName;
        DeviceIo::save(_sFileName,_pDevice);
        _bMustSave=false;
        device_changed(0);
    }
}
//////////////////////////////////////////////////////////////////////////////
void MainWindow::on_actionClose_triggered()
{
    if (ask_save_and_action()==false)
        return;

    clear_device();
    device_changed(0);
}
//////////////////////////////////////////////////////////////////////////////
bool MainWindow::load_file(string sFile)
{
    _sFileName="";
    _bMustSave=false;
    clear_device();
    Device* pDevice=DeviceIo::load(sFile);
    if (pDevice==0)
    {
        QMessageBox m;
        m.setText("Unable to Open file!!");
        m.exec();
        return false;
    }

    delete _pDevice;
    _pDevice=pDevice;
    _sFileName=sFile;
    device_changed(0);
    return true;
}
//////////////////////////////////////////////////////////////////////////////
void MainWindow::device_changed(void* pSender)
{
    qDebug("MainWindow::device_changed");

    if (pSender!=0)
    {
        _bMustSave=true;
    }

    if (pSender!=(void*)_pDockScatterPlot)
    {
        _pDockScatterPlot->device_changed(_pDevice);
    }

    if (pSender!=(void*)_pDockSurfacesData)
    {
        _pDockSurfacesData->device_changed(_pDevice);
    }

    if (pSender!=(void*)_pDockLightProperties)
    {
        _pDockLightProperties->device_changed(_pDevice);
    }

    if (pSender!=(void*)_pDockAutofocus)
    {
        _pDockAutofocus->device_changed(_pDevice);
    }

    if (pSender!=(void*)_pFrameSideView)
    {
        _pFrameSideView->device_changed(_pDevice);
    }

    //maj le titre
    if (_bMustSave==true)
        setWindowTitle(QString(("QtLens [loaded"+_sFileName+"] *").c_str()));
    else
    {
        if (_pDevice!=0)
            setWindowTitle(QString(("QtLens [loaded"+_sFileName+"]").c_str()));
        else
            setWindowTitle("QtLens");
    }
}
//////////////////////////////////////////////////////////////////////////////
